#define __PROG_TYPES_COMPAT__

#define F_CPU	16000000

#include <avr/pgmspace.h>
#include <avr/io.h>

#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>

#include "lcd.h"
#include "gfx.h"

#define CPU			0
#define HUMAN		1

#define NO_SYMBOL	0
#define X_SYMBOL	1
#define O_SIMBOL	2

#define UP			0
#define DOWN		1
#define LEFT		2
#define RIGHT		3
#define OK			4

#define BUTTON_REG	DDRA
#define BUTTON_PORT	PORTA
#define BUTTON_PIN	PINA

#define BUTTON_UP		PA2
#define BUTTON_DOWN		PA3
#define BUTTON_LEFT		PA0
#define BUTTON_RIGHT	PA4
#define BUTTON_OK		PA1

static int8_t player = CPU;
static int8_t won = 0, lost = 0, draw = 0;
static int8_t last_button = 0;
static int8_t cpu_symbol, human_symbol;

struct cell {
	int8_t symbol;
	int8_t x, y;
	int8_t w, h;
};

static struct cell board[3][3];
int8_t c_x, c_y;

void board_init(int16_t x, int16_t y,
		int16_t width, int16_t height)
{
	int16_t cell_width = width / 3;
	int16_t cell_height = height / 3;
	
	int8_t i, j;
	
	board[0][0].symbol = NO_SYMBOL;
	board[0][0].x = x;
	board[0][0].y = y + 2 * cell_height;
	
	board[0][1].symbol = NO_SYMBOL;
	board[0][1].x = x;
	board[0][1].y = y + cell_height;
	
	board[0][2].symbol = NO_SYMBOL;
	board[0][2].x = x;
	board[0][2].y = y;
	
	board[1][0].symbol = NO_SYMBOL;
	board[1][0].x = x + cell_height;
	board[1][0].y = y + 2 * cell_height;
	
	board[1][1].symbol = NO_SYMBOL;
	board[1][1].x = x + cell_height;
	board[1][1].y = y + cell_height;
	
	board[1][2].symbol = NO_SYMBOL;
	board[1][2].x = x + cell_height;
	board[1][2].y = y;
	
	board[2][0].symbol = NO_SYMBOL;
	board[2][0].x = x + 2 * cell_height;
	board[2][0].y = y + 2 * cell_height;
	
	board[2][1].symbol = NO_SYMBOL;
	board[2][1].x = x + 2 * cell_height;
	board[2][1].y = y + cell_height;
	
	board[2][2].symbol = NO_SYMBOL;
	board[2][2].x = x + 2 * cell_height;
	board[2][2].y = y;
	
	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++) {
			board[i][j].w = cell_width;
			board[i][j].h = cell_height;
		}
	
	c_x = 0;
	c_y = 0;
}

static void cell_border_set(int16_t x, int16_t y,
			    int16_t width, int16_t height,
			    uint16_t color)
{
	GFX_draw_rect(x + 1, y + 1, width - 2, height - 2, color, 0);
	GFX_draw_rect(x + 2, y + 2, width - 4, height - 4, color, 0);
}

static void cell_symbol_set(int16_t x, int16_t y,
			    int16_t width, int16_t height,
			    uint16_t color, uint16_t bg, int8_t symbol)
{
	GFX_draw_rect(x + 3, y + 3, width - 5, height - 5, bg, 1);

	switch (symbol) {
	case X_SYMBOL:
		GFX_draw_line(x + 6, y + 6,
			      x + width - 7, y + height - 7,
			      color);
		GFX_draw_line(x + 6, y + height - 7,
			      x + width - 7, y + 6,
			      color);
		break;
	case O_SIMBOL:
		GFX_draw_circle(x + width / 2, y + height / 2,
				width / 2 - 5, color, 0);		
		break;
	default:
		break;
	}
}

static void cell_highlight(struct cell c, uint16_t color, uint16_t bg)
{	
	cell_symbol_set(c.x, c.y, c.w, c.h, ST7735_WHITE, color, c.symbol);

	_delay_ms(200);

	cell_symbol_set(c.x, c.y, c.w, c.h, ST7735_WHITE, bg, c.symbol);
}

static void move(int8_t direction)
{
	struct cell c;

	c = board[c_x][c_y];
	cell_border_set(c.x, c.y, c.w, c.h, ST7735_BLACK);

	switch (direction) {
	case UP:
		if (c_x == 0)
			c_x = 2;
		else
			c_x--;

		break;
	case DOWN:
		if (c_x == 2)
			c_x = 0;
		else
			c_x++;

		break;
	case LEFT:
		if (c_y == 0)
			c_y = 2;
		else
			c_y--;

		break;
	case RIGHT:
		if (c_y == 2)
			c_y = 0;
		else
			c_y++;

		break;
	default:
		break;
	}
	
	c = board[c_x][c_y];
	cell_border_set(c.x, c.y, c.w, c.h, ST7735_WHITE);
}

static void button_set(uint8_t button)
{
	BUTTON_REG &= ~(1 << button);
	BUTTON_PORT |= (1 << button);
}

uint8_t button_press(uint8_t button)
{
	if (!(BUTTON_PIN & (1 << button))) {
		_delay_ms(50);
		
		if (!(BUTTON_PIN & (1 << button))) {
			_delay_ms(50);
			
			if (!(BUTTON_PIN & (1 << button)))
				return 1;
		}
	}
	
	return 0;
}

static void draw_welcome(int16_t x, int16_t y)
{
	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y);
	GFX_print_str("Welcome");
}

static void draw_goodbye(int16_t x, int16_t y)
{
	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y);
	GFX_print_str("Goodbye!");
}

static void draw_title(int16_t x, int16_t y)
{
	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y);
	GFX_print_str("Tic-tac-toe");
}

static void draw_hud(int16_t x, int16_t y)
{
	char games_won[10], games_lost[10]; 

	GFX_draw_rect(y, GFX_get_height() - 50, 90, 50, ST7735_BLACK, 1);

	GFX_set_text_size(1);
	GFX_set_rotation(1);
	
	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y);
	GFX_print_str("Games");
	
	sprintf(games_won, "Won:  %d", won);
	GFX_set_text_color(ST7735_GREEN);
	GFX_set_cursor(x, y + 10);
	GFX_print_str(games_won);
	
	sprintf(games_lost, "Lost: %d", lost);
	GFX_set_text_color(ST7735_RED);
	GFX_set_cursor(x, y + 20);
	GFX_print_str(games_lost);
	
	sprintf(games_lost, "Draw: %d", draw);
	GFX_set_text_color(ST7735_YELLOW);
	GFX_set_cursor(x, y + 30);
	GFX_print_str(games_lost);
	
	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y + 50);
	GFX_print_str("Moving");
	
	GFX_set_text_color(ST7735_MAGENTA);
	GFX_set_cursor(x, y + 60);
	
	if (player == CPU)
		GFX_print_str("CPU");
	else
		GFX_print_str("YOU");
}

static void draw_grid(int16_t x, int16_t y,
		      int16_t width, int16_t height,
		      uint16_t color)
{
	int16_t cell_width = width / 3;
	int16_t cell_height = height / 3;
	
	int16_t i, j;
	
	GFX_draw_rect(x, y, width, height, ST7735_BLACK, 1);
	
	for (i = x; i < x + width; i += cell_width)
		for (j = y; j < y + height; j += cell_height)
			GFX_draw_rect(i, j, cell_width, cell_height, color, 0);
}

static int won_game(int symb)
{
	int8_t i;
	
	for (i = 0; i < 3; i++) {
		if (board[i][0].symbol == board[i][1].symbol && board[i][1].symbol == board[i][2].symbol && board[i][2].symbol == symb) {
			cell_highlight(board[i][0], ST7735_GREEN, ST7735_BLACK);
			cell_highlight(board[i][1], ST7735_GREEN, ST7735_BLACK);
			cell_highlight(board[i][2], ST7735_GREEN, ST7735_BLACK);
			return 1;
		}
		
		if (board[0][i].symbol == board[1][i].symbol && board[1][i].symbol == board[2][i].symbol && board[2][i].symbol == symb) {
			cell_highlight(board[0][i], ST7735_GREEN, ST7735_BLACK);
			cell_highlight(board[1][i], ST7735_GREEN, ST7735_BLACK);
			cell_highlight(board[2][i], ST7735_GREEN, ST7735_BLACK);
			return 1;
		}
	}
	
	if (board[0][0].symbol == board[1][1].symbol && board[1][1].symbol == board[2][2].symbol && board[2][2].symbol == symb) {
		cell_highlight(board[0][0], ST7735_GREEN, ST7735_BLACK);
		cell_highlight(board[1][1], ST7735_GREEN, ST7735_BLACK);
		cell_highlight(board[2][2], ST7735_GREEN, ST7735_BLACK);
		return 1;
	}
	
	if (board[0][2].symbol == board[1][1].symbol && board[1][1].symbol == board[2][0].symbol && board[2][0].symbol == symb) {
		cell_highlight(board[0][2], ST7735_GREEN, ST7735_BLACK);
		cell_highlight(board[1][1], ST7735_GREEN, ST7735_BLACK);
		cell_highlight(board[2][0], ST7735_GREEN, ST7735_BLACK);
		return 1;
	}
	
	return 0;
}

static int full_board()
{
	int8_t i, j, full = 1;

	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			if (board[i][j].symbol == NO_SYMBOL) {
				full = 0;
				break;
			}
			
	return full;
}

static int game_over()
{
	if (!won_game(cpu_symbol) && !won_game(human_symbol) && !full_board())
		return 0;
	
	if (won_game(cpu_symbol))
		lost++;
	else if (won_game(human_symbol))
		won++;
	else if (full_board())
		draw++;

	return 1;
}

static void cpu_move()
{	
	int tries[24][3][2] = {{{0, 0}, {0, 1}, {0, 2}}, {{0, 0}, {0, 2}, {0, 1}}, {{0, 0}, {1, 0}, {2, 0}}, 
							{{0, 0}, {1, 1}, {2, 2}}, {{0, 0}, {2, 0}, {1, 0}}, {{0, 0}, {2, 2}, {1, 1}},
							{{0, 1}, {0, 2}, {0, 0}}, {{0, 1}, {1, 1}, {2, 1}}, {{0, 1}, {2, 1}, {1, 1}}, 
							{{0, 2}, {1, 1}, {2, 0}}, {{0, 2}, {1, 2}, {2, 2}}, {{0, 2}, {2, 0}, {1, 1}},
							{{0, 2}, {2, 2}, {1, 2}}, {{1, 0}, {1, 1}, {1, 2}}, {{1, 0}, {1, 2}, {1, 1}}, 
							{{1, 0}, {2, 0}, {0, 0}}, {{1, 1}, {1, 2}, {1, 0}}, {{1, 1}, {2, 0}, {0, 2}},
							{{1, 1}, {2, 1}, {0, 1}}, {{1, 1}, {2, 2}, {0, 0}}, {{1, 2}, {2, 2}, {0, 2}}, 
							{{2, 0}, {2, 1}, {2, 2}}, {{2, 0}, {2, 2}, {2, 1}}, {{2, 1}, {2, 2}, {2, 0}}};

	int8_t i, j, n = 0, index_x, index_y, try = 0;

	for (i = 0; i < 24; i++) {
		int8_t xi = tries[i][0][0];
		int8_t xj = tries[i][0][1];
		int8_t yi = tries[i][1][0];
		int8_t yj = tries[i][1][1];
		int8_t zi = tries[i][2][0];
		int8_t zj = tries[i][2][1];
		
		if (board[xi][xj].symbol == board[yi][yj].symbol && 
				board[xi][xj].symbol != NO_SYMBOL && 
					board[zi][zj].symbol == NO_SYMBOL) {
			index_x = zi;
			index_y = zj;
			try = 1;
			break;
		}
	}

	if (!try) {
		int free_cells[9];
		for (i = 0; i < 3; i++)
			for (j = 0; j < 3; j++)
				if (board[i][j].symbol == NO_SYMBOL)
					free_cells[n++] = i * 3 + j;
		
		int8_t index = free_cells[rand() % n];
		index_x = index / 3;
		index_y = index % 3;
	}

	board[index_x][index_y].symbol = cpu_symbol;
	struct cell c = board[index_x][index_y];
	cell_symbol_set(c.x, c.y, c.w, c.h, ST7735_WHITE, ST7735_BLACK, cpu_symbol);
}

static int set_cpu_symbol()
{
	return rand() % 2 + 1;
}

static int set_human_symbol()
{
	return (cpu_symbol == X_SYMBOL) ? O_SIMBOL : X_SYMBOL;
}

int main(void)
{
	LCD_init();
	GFX_init(128, 160);
	
	LCD_fill_screen(ST7735_BLACK);
	draw_welcome(40, 50);
	_delay_ms(2000);
	LCD_fill_screen(ST7735_BLACK);
	
	draw_title(15, 5);
	
	button_set(BUTTON_UP);
	button_set(BUTTON_DOWN);
	button_set(BUTTON_LEFT);
	button_set(BUTTON_RIGHT);
	button_set(BUTTON_OK);
	
	while (1) {

		last_button = BUTTON_UP;

		cpu_symbol = set_cpu_symbol();
		human_symbol = set_human_symbol();
		
		if (human_symbol == X_SYMBOL)
			player = HUMAN;
		else
			player = CPU;

		draw_hud(5, 30);
		draw_grid(30, 5, 90, 90, ST7735_WHITE);

		board_init(30, 5, 90, 90);
		cell_border_set(board[0][0].x, board[0][0].y,
			        board[0][0].w, board[0][0].h,
			        ST7735_WHITE);
		
		while (!game_over()) {
			
			if (player == CPU) {
				_delay_ms(1000);
				cpu_move(cpu_symbol);
				player = HUMAN;
				draw_hud(5, 30);
			} else {
				if (button_press(BUTTON_UP)) {
					last_button = UP;
					move(UP);
				} else if (button_press(BUTTON_DOWN)) {
					last_button = DOWN;
					move(DOWN);
				} else if (button_press(BUTTON_LEFT)) {
					last_button = LEFT;
					move(LEFT);
				} else if (button_press(BUTTON_RIGHT)) {
					last_button = RIGHT;
					move(RIGHT);
				} else if (button_press(BUTTON_OK)) {
					if (last_button == OK)
						goto exit;

					last_button = OK;
					
					if (board[c_x][c_y].symbol != NO_SYMBOL) {
						cell_highlight(board[c_x][c_y], ST7735_RED, ST7735_BLACK);						
						_delay_ms(1000);
						continue;
					}
					
					board[c_x][c_y].symbol = human_symbol;
					struct cell c = board[c_x][c_y];

					cell_symbol_set(c.x, c.y,
							c.w, c.h,
							ST7735_WHITE, ST7735_BLACK,
							c.symbol);
					
					player = CPU;
					draw_hud(5, 30);
				}
			}
		}

		_delay_ms(2000);
	}

exit:

	LCD_fill_screen(ST7735_BLACK);
	draw_goodbye(35, 50);
	_delay_ms(2000);	
	LCD_fill_screen(ST7735_BLACK);

	LCD_end();

	return 0;
}
